from galaxy.tool_util.unittest_utils import t_data_downloader_for
from galaxy.tool_util.verify import verify


def test_writes_output_to_target_directory(tmp_path):
    item_label = "my test case"
    output_content = b"expected"
    filename = "my_output.txt"

    assert not (tmp_path / filename).exists()
    verify(
        item_label,
        output_content,
        attributes={},
        filename=filename,
        keep_outputs_dir=str(tmp_path),
        get_filecontent=t_data_downloader_for(output_content),
    )
    assert (tmp_path / filename).exists()
    assert (tmp_path / filename).open("rb").read() == output_content


def test_no_writes_output_to_target_directory_on_contains(tmp_path):
    item_label = "my test case"
    output_content = b"expected"
    attributes = {
        "compare": "contains",
    }
    filename = "my_output.txt"

    get_filecontent = t_data_downloader_for(b"xpecte")
    assert not (tmp_path / filename).exists()
    verify(
        item_label,
        output_content,
        attributes=attributes,
        filename=filename,
        keep_outputs_dir=str(tmp_path),
        get_filecontent=get_filecontent,
    )
    # Consider switching this to not update - see discussion on
    # https://github.com/galaxyproject/galaxy/pull/14661
    assert (tmp_path / filename).exists()


def test_sim_size(tmp_path):
    item_label = "my test case"
    output_content = b"expected"
    attributes = {
        "compare": "sim_size",
        "delta": 2,
    }
    filename = "my_output.txt"

    get_filecontent = t_data_downloader_for(b"xpected")
    assert not (tmp_path / filename).exists()
    verify(
        item_label,
        output_content,
        attributes=attributes,
        filename=filename,
        keep_outputs_dir=str(tmp_path),
        get_filecontent=get_filecontent,
    )
    assert (tmp_path / filename).exists()
    assert (tmp_path / filename).open("rb").read() == b"expected"


def test_sim_size_failure_still_updates(tmp_path):
    item_label = "my test case"
    output_content = b"expected"
    attributes = {
        "compare": "sim_size",
        "delta": 2,
    }
    filename = "my_output.txt"

    get_filecontent = t_data_downloader_for(b"ected")
    assert not (tmp_path / filename).exists()
    assertion_error = None
    try:
        verify(
            item_label,
            output_content,
            attributes=attributes,
            filename=filename,
            keep_outputs_dir=str(tmp_path),
            get_filecontent=get_filecontent,
        )
    except AssertionError as ae:
        assertion_error = ae

    assert assertion_error
    assert (tmp_path / filename).exists()
    assert (tmp_path / filename).open("rb").read() == b"expected"



import os.path

import pytest

from galaxy.tool_util.deps.mulled.mulled_build import (
    base_image_for_targets,
    build_target,
    DEFAULT_BASE_IMAGE,
    DEFAULT_EXTENDED_BASE_IMAGE,
    InvolucroContext,
    mull_targets,
    target_str_to_targets,
)
from galaxy.tool_util.deps.mulled.util import CondaInDockerContext
from ..util import external_dependency_management


@pytest.mark.parametrize(
    "target,version,base_image",
    [
        ("mzmine", None, DEFAULT_EXTENDED_BASE_IMAGE),
        ("qiime", "1.9.1", DEFAULT_EXTENDED_BASE_IMAGE),
        ("samtools", None, DEFAULT_BASE_IMAGE),
    ],
)
@external_dependency_management
def test_base_image_for_targets(target, version, base_image):
    target = build_target(target, version=version)
    conda_context = CondaInDockerContext()
    assert base_image_for_targets([target], conda_context) == base_image


@pytest.mark.parametrize("use_mamba", [False, True])
@external_dependency_management
def test_mulled_build_files_cli(use_mamba: bool, tmpdir) -> None:
    singularity_image_dir = tmpdir.mkdir("singularity image dir")
    target = build_target("zlib", version="1.2.13", build="h166bdaf_4")
    involucro_context = InvolucroContext(involucro_bin=os.path.join(tmpdir, "involucro"))
    exit_code = mull_targets(
        [target],
        determine_base_image=False,
        involucro_context=involucro_context,
        command="build-and-test",
        singularity=True,
        use_mamba=use_mamba,
        singularity_image_dir=singularity_image_dir,
    )
    assert exit_code == 0
    assert singularity_image_dir.join("zlib:1.2.13--h166bdaf_4").exists()


def test_target_str_to_targets():
    target_str = "samtools=1.3.1--4,bedtools=2.22"
    targets = target_str_to_targets(target_str)
    assert (targets[0].package, targets[0].version, targets[0].build) == ("samtools", "1.3.1", "4")
    assert (targets[1].package, targets[1].version, targets[1].build) == ("bedtools", "2.22", None)



import json
from contextlib import contextmanager
from os import environ
from typing import (
    Any,
    List,
)

import pytest

external_dependency_management = pytest.mark.external_dependency_management


def dict_verify_each(target_dict: dict, expectations: List[Any]):
    assert_json_encodable(target_dict)
    for path, expectation in expectations:
        exception = target_dict.get("exception")
        assert not exception, f"Test failed to generate with exception {exception}"
        dict_verify(target_dict, path, expectation)


def dict_verify(target_dict: dict, expectation_path: List[Any], expectation: Any):
    rest = target_dict
    for path_part in expectation_path:
        rest = rest[path_part]
    assert rest == expectation, f"{rest} != {expectation} for {expectation_path}"


def assert_json_encodable(as_dict: dict):
    json.dumps(as_dict)


@contextmanager
def modify_environ(values, keys_to_remove=None):
    """
    Modify the environment for a test, adding/updating values in dict `values` and
    removing any environment variables mentioned in list `keys_to_remove`.
    """
    old_environ = environ.copy()
    try:
        if values:
            environ.update(values)
        if keys_to_remove:
            for key in keys_to_remove:
                if key in environ:
                    del environ[key]
        yield
    finally:
        environ.clear()
        environ.update(old_environ)



"""Tool test parsing to dicts logic."""

import os
from typing import (
    Any,
    List,
)

from pytest import skip

from galaxy.tool_util.parser.factory import get_tool_source
from galaxy.tool_util.unittest_utils import functional_test_tool_path
from galaxy.tool_util.verify.parse import parse_tool_test_descriptions
from galaxy.util import (
    galaxy_directory,
    in_packages,
)
from galaxy.util.unittest import TestCase
from .util import dict_verify_each

# Not the whole response, just some keys and such to test...
SIMPLE_CONSTRUCTS_EXPECTATIONS_0 = [
    (["inputs", "booltest"], [True]),
    (["inputs", "inttest"], ["12456"]),
    (["inputs", "floattest"], ["6.789"]),
    (["inputs", "p1|p1use"], [True]),
    (["inputs", "files_0|file"], ["simple_line.txt"]),
    (["outputs", 0, "name"], "out_file1"),
    (["required_files", 0, 0], "simple_line.txt"),
    (["required_files", 0, 1, "value"], "simple_line.txt"),
]
# this test case covers specifying boolean parameters by string truevalue/falsevalue
SIMPLE_CONSTRUCTS_EXPECTATIONS_1 = [
    (["inputs", "booltest"], [True]),
    (["inputs", "p1|p1use"], [True]),
    (["inputs", "files_0|file"], ["simple_line.txt"]),
    (["outputs", 0, "name"], "out_file1"),
]
SECTION_EXPECTATIONS = [
    (["inputs", "int|inttest"], ["12456"]),
    (["inputs", "float|floattest"], ["6.789"]),
]
MIN_REPEAT_EXPECTATIONS = [
    (["inputs", "queries_0|input"], ["simple_line.txt"]),
    (["inputs", "queries_1|input"], ["simple_line.txt"]),
    (["inputs", "queries2_0|input2"], ["simple_line_alternative.txt"]),
]
DBKEY_FILTER_INPUT_EXPECTATIONS = [
    (["inputs", "inputs"], ["simple_line.txt"]),
    (["inputs", "index"], ["hg19_value"]),
    (["required_files", 0, 1, "dbkey"], "hg19"),
    (["required_data_tables", 0], "test_fasta_indexes"),
]
COLLECTION_TYPE_SOURCE_EXPECTATIONS = [
    (["inputs", "input_collect", "model_class"], "TestCollectionDef"),
    (["inputs", "input_collect", "collection_type"], "list"),
]
BIGWIG_TO_WIG_EXPECTATIONS = [
    (["inputs", "chrom"], "chr21"),
]
RECORD_TWO_FILES_EXPECTATIONS = [
    (["inputs", "f1", "model_class"], "TestCollectionDef"),
    (["inputs", "f1", "collection_type"], "record"),
    (["inputs", "f1", "fields", 0, "type"], "File"),
    (["inputs", "f1", "fields", 0, "name"], "parent"),
    (["inputs", "f1", "fields", 1, "type"], "File"),
]


class TestTestParsing(TestCase):
    def _parse_tests(self):
        return parse_tool_test_descriptions(self.tool_source)

    def _init_tool_for_path(self, path):
        tool_source = get_tool_source(path)
        self.tool_source = tool_source

    def test_maxseconds_not_filled_with_default(self):
        self._init_tool_for_path(functional_test_tool_path("simple_constructs.xml"))
        test_dicts = list(self._parse_tests())
        assert test_dicts[0].maxseconds is None
        assert "maxseconds" not in test_dicts[0].to_dict()

    def test_maxseconds_parsed(self):
        self._init_tool_for_path(functional_test_tool_path("maxseconds.xml"))
        test_def = next(iter(self._parse_tests()))
        assert test_def.maxseconds == 5
        test_dict = test_def.to_dict()
        print(test_dict)
        assert "maxseconds" in test_dict
        assert test_dict["maxseconds"] == 5

    def test_simple_state_parsing(self):
        self._init_tool_for_path(functional_test_tool_path("simple_constructs.xml"))
        test_dicts = self._parse_tests()
        self._verify_each(test_dicts[0].to_dict(), SIMPLE_CONSTRUCTS_EXPECTATIONS_0)
        self._verify_each(test_dicts[1].to_dict(), SIMPLE_CONSTRUCTS_EXPECTATIONS_1)

    def test_section_state_parsing(self):
        self._init_tool_for_path(functional_test_tool_path("section.xml"))
        test_dicts = self._parse_tests()
        # without and with section tags in the tests...
        self._verify_each(test_dicts[0].to_dict(), SECTION_EXPECTATIONS)
        self._verify_each(test_dicts[1].to_dict(), SECTION_EXPECTATIONS)

    def test_repeat_state_parsing(self):
        self._init_tool_for_path(functional_test_tool_path("min_repeat.xml"))
        test_dicts = self._parse_tests()
        # without and with section tags in the tests...
        self._verify_each(test_dicts[0].to_dict(), MIN_REPEAT_EXPECTATIONS)

    def test_dynamic_options_data_table_parsing(self):
        self._init_tool_for_path(functional_test_tool_path("dbkey_filter_input.xml"))
        test_dicts = self._parse_tests()
        self._verify_each(test_dicts[0].to_dict(), DBKEY_FILTER_INPUT_EXPECTATIONS)

    def test_collection_type_source_parsing(self):
        self._init_tool_for_path(functional_test_tool_path("collection_type_source.xml"))
        test_dicts = self._parse_tests()
        self._verify_each(test_dicts[0].to_dict(), COLLECTION_TYPE_SOURCE_EXPECTATIONS)

    def test_unqualified_access_disabled_in_24_2(self):
        self._init_tool_for_path(functional_test_tool_path("deprecated/simple_constructs_24_2.xml"))
        test_dicts = self._parse_tests()
        test_0 = test_dicts[0].to_dict()
        assert test_0["error"] is True

    def test_field_collection_inputs(self):
        self._init_tool_for_path(functional_test_tool_path("collection_record_test_two_files.xml"))
        test_dicts = self._parse_tests()
        test_0 = test_dicts[0].to_dict()
        self._verify_each(test_0, RECORD_TWO_FILES_EXPECTATIONS)

    def test_bigwigtowig_converter(self):
        # defines
        if in_packages():
            skip(
                "skipping this mode for now - we need a framework test tool that skips names and just specifies arguments"
            )
        tool_path = os.path.join(
            galaxy_directory(), "lib", "galaxy", "datatypes", "converters", "bigwig_to_wig_converter.xml"
        )
        self._init_tool_for_path(tool_path)
        test_dicts = self._parse_tests()
        self._verify_each(test_dicts[1].to_dict(), BIGWIG_TO_WIG_EXPECTATIONS)

    def _verify_each(self, target_dict: dict, expectations: List[Any]):
        exception = target_dict.get("exception")
        assert not exception, f"Test failed to generate with exception {exception}"
        dict_verify_each(target_dict, expectations)



from subprocess import CalledProcessError

import pytest

from galaxy.tool_util.deps.container_classes import DOCKER_CONTAINER_TYPE
from galaxy.tool_util.deps.container_resolvers.mulled import (
    CachedMulledDockerContainerResolver,
    CachedMulledSingularityContainerResolver,
    MulledDockerContainerResolver,
)
from galaxy.tool_util.deps.containers import ContainerRegistry
from galaxy.tool_util.deps.dependencies import (
    AppInfo,
    ToolInfo,
)
from galaxy.tool_util.deps.requirements import ToolRequirement

SINGULARITY_IMAGES = (
    "foo:1.0--bar",
    "baz:2.22",
    "mulled-v2-fe8a3b846bc50d24e5df78fa0b562c43477fe9ce:9f946d13f673ab2903cb0da849ad42916d619d18-0",
)


@pytest.fixture
def appinfo() -> AppInfo:
    return AppInfo(
        involucro_auto_init=True,
        enable_mulled_containers=True,
        container_image_cache_path=".",
    )


@pytest.fixture
def container_registry():
    app_info = AppInfo(
        involucro_auto_init=True,
        enable_mulled_containers=True,
        container_image_cache_path=".",
    )
    return ContainerRegistry(app_info)


def test_container_registry(container_registry, mocker):
    mocker.patch("galaxy.tool_util.deps.mulled.util._get_namespace", return_value=["samtools"])
    tool_info = ToolInfo(requirements=[ToolRequirement(name="samtools", version="1.10", type="package")])
    container_description = container_registry.find_best_container_description(
        [DOCKER_CONTAINER_TYPE],
        tool_info,
        install=False,
    )
    assert container_description.type == "docker"
    assert "samtools:1.10" in container_description.identifier


def test_docker_container_resolver_detects_docker_cli_absent(appinfo, mocker):
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.which", return_value=None)
    resolver = CachedMulledDockerContainerResolver(appinfo)
    assert resolver._cli_available is False


def test_docker_container_resolver_detects_docker_cli(appinfo, mocker):
    """
    - CachedMulledDockerContainerResolver properly detects present docker binary
    """
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled", return_value="/bin/docker")
    resolver = CachedMulledDockerContainerResolver(appinfo)
    assert resolver.cli_available


def test_cached_docker_container_docker_cli_absent_resolve(appinfo, mocker) -> None:
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.which", return_value=None)
    resolver = CachedMulledDockerContainerResolver(appinfo)
    assert resolver.cli_available is False
    assert resolver.resolve(enabled_container_types=[], tool_info=ToolInfo()) is None


def test_docker_container_docker_cli_absent_resolve(appinfo, mocker):
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.which", return_value=None)
    resolver = MulledDockerContainerResolver(appinfo)
    assert resolver.cli_available is False
    requirement = ToolRequirement(name="samtools", version="1.10", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    mocker.patch(
        "galaxy.tool_util.deps.container_resolvers.mulled.targets_to_mulled_name",
        return_value="samtools:1.10--h2e538c0_3",
    )
    container_description = resolver.resolve(enabled_container_types=["docker"], tool_info=tool_info)
    assert container_description
    assert container_description.type == "docker"
    assert container_description.identifier == "quay.io/biocontainers/samtools:1.10--h2e538c0_3"


def test_docker_container_docker_cli_exception_resolve(appinfo, mocker):
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.which", return_value="/bin/docker")
    resolver = MulledDockerContainerResolver(appinfo)
    assert resolver.cli_available is True
    requirement = ToolRequirement(name="samtools", version="1.10", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    mocker.patch(
        "galaxy.tool_util.deps.container_resolvers.mulled.targets_to_mulled_name",
        return_value="samtools:1.10--h2e538c0_3",
    )
    mocker.patch(
        "galaxy.tool_util.deps.container_resolvers.mulled.docker_cached_container_description",
        side_effect=CalledProcessError(1, "bla"),
    )
    container_description = resolver.resolve(enabled_container_types=["docker"], tool_info=tool_info, install=True)
    assert resolver.cli_available is True
    assert container_description
    assert container_description.type == "docker"
    assert container_description.identifier == "quay.io/biocontainers/samtools:1.10--h2e538c0_3"


def test_cached_singularity_container_resolver_uncached(mocker):
    mocker.patch("os.listdir", return_value=SINGULARITY_IMAGES)
    mocker.patch("os.path.exists", return_value=True)
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.safe_makedirs")
    resolver = CachedMulledSingularityContainerResolver(app_info=mocker.Mock(container_image_cache_path="/"))
    requirement = ToolRequirement(name="foo", version="1.0", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    container_description = resolver.resolve(enabled_container_types=["singularity"], tool_info=tool_info)
    assert container_description
    assert container_description.type == "singularity"
    assert container_description.identifier == "/singularity/mulled/foo:1.0--bar"


def test_cached_singularity_container_resolver_dir_mtime_cached(mocker):
    mocker.patch("os.listdir", return_value=SINGULARITY_IMAGES)
    mocker.patch("os.path.exists", return_value=True)
    mocker.patch("galaxy.tool_util.deps.container_resolvers.mulled.safe_makedirs")
    mocker.patch("os.stat", return_value=mocker.Mock(st_mtime=42))
    resolver = CachedMulledSingularityContainerResolver(
        app_info=mocker.Mock(container_image_cache_path="/"), cache_directory_cacher_type="dir_mtime"
    )
    requirement = ToolRequirement(name="baz", version="2.22", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    container_description = resolver.resolve(enabled_container_types=["singularity"], tool_info=tool_info)
    assert container_description
    assert container_description.type == "singularity"
    assert container_description.identifier == "/singularity/mulled/baz:2.22"
    requirement = ToolRequirement(name="foo", version="1.0", type="package")
    tool_info.requirements.append(requirement)
    container_description = resolver.resolve(enabled_container_types=["singularity"], tool_info=tool_info)
    assert container_description
    assert container_description.type == "singularity"
    assert (
        container_description.identifier
        == "/singularity/mulled/mulled-v2-fe8a3b846bc50d24e5df78fa0b562c43477fe9ce:9f946d13f673ab2903cb0da849ad42916d619d18-0"
    )



from galaxy.tool_util.ontologies.ontology_data import expand_ontology_data
from .test_parsing import (
    get_test_tool_source,
    TOOL_YAML_1,
)

TOOL_YAML_2 = """
name: "Bowtie Mapper"
class: GalaxyTool
id: bowtie
version: 1.0.2
description: "The Bowtie Mapper"
xrefs:
  - type: bio.tools
    value: bwa
command: "bowtie --map-the-stuff"
outputs:
  out1:
    format: bam
    from_work_dir: out1.bam
edam_operations:
 - operation_0335
edam_topics:
 - topic_0102
inputs:
  - name: input1
    type: integer
"""

TOOL_YAML_3 = """
name: "Bowtie Mapper"
class: GalaxyTool
id: sort1
version: 1.0.2
description: "The Bowtie Mapper"
xrefs:
  - type: bio.tools
    value: bwa
command: "bowtie --map-the-stuff"
outputs:
  out1:
    format: bam
    from_work_dir: out1.bam
inputs:
  - name: input1
    type: integer
"""

TOOL_YAML_NO_EXPLICIT_XREFS = """
name: "Bowtie Mapper"
class: GalaxyTool
id: sort1
version: 1.0.2
description: "The Bowtie Mapper"
command: "bowtie --map-the-stuff"
outputs:
  out1:
    format: bam
    from_work_dir: out1.bam
inputs:
  - name: input1
    type: integer
"""


def test_parse_edam_empty():
    test_source = get_test_tool_source(source_file_name="testtool.yml", source_contents=TOOL_YAML_1)
    ontology_data = expand_ontology_data(test_source, ["bowtie"], None)
    assert ontology_data.edam_operations == []
    assert ontology_data.edam_topics == []


def test_parse_edam_direct():
    test_source = get_test_tool_source(source_file_name="testtool.yml", source_contents=TOOL_YAML_2)
    ontology_data = expand_ontology_data(test_source, ["bowtie"], None)
    assert ontology_data.edam_operations == ["operation_0335"]
    assert ontology_data.edam_topics == ["topic_0102"]


def test_parse_edam_mapping_operations_legacy():
    test_source = get_test_tool_source(source_file_name="testtool.yml", source_contents=TOOL_YAML_3)
    ontology_data = expand_ontology_data(test_source, ["sort1"], None)
    assert ontology_data.edam_operations == ["operation_3802"]
    assert ontology_data.edam_topics == []


def test_parse_biotools_default_mapping():
    test_source = get_test_tool_source(source_file_name="testtool.yml", source_contents=TOOL_YAML_NO_EXPLICIT_XREFS)
    ontology_data = expand_ontology_data(test_source, ["cheetah_problem_unbound_var_input"], None)
    assert ontology_data.xrefs[0]["type"] == "bio.tools"
    assert ontology_data.xrefs[0]["value"] == "bwa"

    test_source = get_test_tool_source(source_file_name="testtool.yml", source_contents=TOOL_YAML_NO_EXPLICIT_XREFS)
    ontology_data = expand_ontology_data(test_source, ["rxlr_motifs"], None)
    assert len(ontology_data.xrefs) == 2
    values = [x["value"] for x in ontology_data.xrefs]
    assert "signalp" in values
    assert "hmmer2" in values



import pytest

from galaxy.tool_util.deps.mulled.mulled_search import (
    CondaSearch,
    get_package_hash,
    GitHubSearch,
    QuaySearch,
    singularity_search,
)
from galaxy.util.unittest_utils import skip_unless_executable
from ..util import external_dependency_management


@external_dependency_management
def test_quay_search():
    t = QuaySearch("biocontainers")
    t.build_index()
    search1 = t.search_repository("adsfasdf", True)
    search2 = t.search_repository("bioconductor-gosemsim", True)
    assert search1 == []
    assert {"version": "2.2.0--0", "package": "bioconductor-gosemsim"} in search2


@external_dependency_management
@skip_unless_executable("conda")
def test_conda_search():
    t = CondaSearch("bioconda")
    search = t.get_json("asdfasdf")
    assert search == []
    search = t.get_json("bioconductor-gosemsim")
    assert all(r["package"] == "bioconductor-gosemsim" for r in search)


@external_dependency_management
def test_github_recipe_present():
    t = GitHubSearch()

    search_string2expected = {
        "adsfasdf": False,
        "bioconductor-gosemsim": True,
    }
    for search_string, expected in search_string2expected.items():
        try:
            is_recipe_present = t.recipe_present(search_string)
        except Exception as e:
            if "API rate limit" in str(e):
                pytest.skip("Hitting GitHub API rate limit")
            raise
        assert is_recipe_present is expected


@external_dependency_management
def test_get_package_hash():
    package_hash1 = get_package_hash(["bamtools", "samtools"], {})
    package_hash2 = get_package_hash(["bamtools", "samtools"], {"bamtools": "2.4.0", "samtools": "1.3.1"})
    package_hash3 = get_package_hash(["abricate", "abyss"], {"abricate": "0.4", "abyss": "2.0.1"})
    assert package_hash1["package_hash"] == "mulled-v2-0560a8046fc82aa4338588eca29ff18edab2c5aa"
    assert package_hash2["version_hash"] == "c17ce694dd57ab0ac1a2b86bb214e65fedef760e"
    assert package_hash2["package_hash"] == "mulled-v2-0560a8046fc82aa4338588eca29ff18edab2c5aa"
    assert package_hash3["version_hash"] == "e21d1262f064e1e01b6b9fad5bea117928f31b38"
    assert package_hash3["package_hash"] == "mulled-v2-cde36934a4704f448af44bf01deeae8d2832ca2e"


@external_dependency_management
def test_singularity_search():
    sing1 = singularity_search("mulled-v2-0560a8046fc82aa4338588eca29ff18edab2c5aa")
    sing1_versions = {result["version"] for result in sing1}
    assert {
        "c17ce694dd57ab0ac1a2b86bb214e65fedef760e-0",
        "f471ba33d45697daad10614c5bd25a67693f67f1-0",
        "fc33176431a4b9ef3213640937e641d731db04f1-0",
    }.issubset(sing1_versions)
    sing2 = singularity_search("mulled-v2-19fa9431f5863b2be81ff13791f1b00160ed0852")
    assert sing2 == []



from galaxy.tool_util.deps.container_volumes import DockerVolume


def test_docker_volume_valid():
    volume_string = "/a:/b:ro"
    docker_volume = DockerVolume.from_str(volume_string)
    assert docker_volume.mode_is_valid
    assert str(docker_volume) == volume_string


def test_docker_volume_valid_host_path_only():
    docker_volume = DockerVolume.from_str("/a:ro")
    assert docker_volume.mode_is_valid
    assert str(docker_volume) == "/a:/a:ro"


def test_docker_volume_not_valid():
    docker_volume = DockerVolume.from_str("/a:/b")
    assert not docker_volume.mode_is_valid



import json
import os
from uuid import uuid4

import yaml

from galaxy.tool_util.cwl import (
    tool_proxy,
    workflow_proxy,
)
from galaxy.tool_util.cwl.parser import (
    _to_cwl_tool_object,
    tool_proxy_from_persistent_representation,
)
from galaxy.tool_util.parser.cwl import CwlToolSource
from galaxy.tool_util.parser.factory import get_tool_source as _get_tool_source


def _functional_test_tool_directory() -> str:
    root_dir = os.getcwd()
    return os.path.join(root_dir, "test/functional/tools")


CWL_TOOLS_DIRECTORY = os.path.abspath(os.path.join(_functional_test_tool_directory(), "cwl_tools"))


def get_tool_source(*args, **kwargs):
    tool_source = _get_tool_source(*args, **kwargs)
    if not isinstance(tool_source, CwlToolSource):
        raise Exception(f"Returned ToolSource class '{type(tool_source)}' was not expected class 'CwlToolSource'")
    return tool_source


def _cwl_tool_path(path: str):
    return os.path.join(CWL_TOOLS_DIRECTORY, path)


def test_tool_proxy():
    """Test that tool proxies load some valid tools correctly."""
    tool_proxy(_cwl_tool_path("v1.0/v1.0/cat1-testcli.cwl"))
    tool_proxy(_cwl_tool_path("v1.0/v1.0/cat3-tool.cwl"))
    tool_proxy(_cwl_tool_path("v1.0/v1.0/env-tool1.cwl"))
    tool_proxy(_cwl_tool_path("v1.0/v1.0/sorttool.cwl"))
    tool_proxy(_cwl_tool_path("v1.0/v1.0/bwa-mem-tool.cwl"))
    tool_proxy(_cwl_tool_path("v1.0/v1.0/parseInt-tool.cwl"))


def test_serialize_deserialize():
    path = _cwl_tool_path("v1.0/v1.0/cat5-tool.cwl")
    tool = tool_proxy(path)
    expected_uuid = tool.uuid
    rep = tool.to_persistent_representation()
    tool = tool_proxy_from_persistent_representation(rep)
    assert tool.uuid == expected_uuid
    tool.job_proxy({"file1": "/moo"}, {})
    with open(path) as f:
        tool_object = yaml.safe_load(f)
    tool_object = json.loads(json.dumps(tool_object))
    tool = _to_cwl_tool_object(tool_object=tool_object, uuid=expected_uuid)
    assert tool.uuid == expected_uuid


def test_serialize_deserialize_workflow_embed():
    # Test inherited hints and requirements from workflow -> tool
    # work here.
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines2-wf.cwl"))
    step_proxies = proxy.step_proxies()
    tool_proxy = step_proxies[0].tool_proxy
    assert tool_proxy.requirements, tool_proxy.requirements


def test_reference_proxies():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines1-wf.cwl"))
    assert len(proxy.tool_reference_proxies()) == 2


def test_subworkflow_parsing():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines10-wf.cwl"))
    assert len(proxy.tool_reference_proxies()) == 2

    assert len(proxy.output_labels) == 1
    assert "count_output" in proxy.output_labels, proxy.output_labels

    galaxy_workflow_dict = proxy.to_dict()
    steps = galaxy_workflow_dict["steps"]
    assert len(steps) == 2  # One input, one subworkflow

    subworkflow_step = steps[1]
    assert subworkflow_step["type"] == "subworkflow"


def test_checks_is_a_tool():
    """Test that tool proxy cannot be created for a workflow."""
    exception = None
    try:
        tool_proxy(_cwl_tool_path("v1.0/v1.0/count-lines1-wf.cwl"))
    except Exception as e:
        exception = e

    assert exception is not None
    assert "CommandLineTool" in str(exception), str(exception)


def test_workflow_of_files_proxy():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines1-wf.cwl"))
    step_proxies = proxy.step_proxies()
    assert len(step_proxies) == 2

    galaxy_workflow_dict = proxy.to_dict()

    assert len(proxy.runnables) == 2

    assert len(galaxy_workflow_dict["steps"]) == 3
    wc_step = galaxy_workflow_dict["steps"][1]
    exp_step = galaxy_workflow_dict["steps"][2]
    assert wc_step["input_connections"]
    assert exp_step["input_connections"]


def test_workflow_embedded_tools_proxy():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines2-wf.cwl"))
    step_proxies = proxy.step_proxies()
    assert len(step_proxies) == 2
    galaxy_workflow_dict = proxy.to_dict()

    assert len(proxy.runnables) == 2

    assert len(galaxy_workflow_dict["steps"]) == 3
    wc_step = galaxy_workflow_dict["steps"][1]
    exp_step = galaxy_workflow_dict["steps"][2]
    assert wc_step["input_connections"]
    assert exp_step["input_connections"]


def test_workflow_scatter():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines3-wf.cwl"))

    step_proxies = proxy.step_proxies()
    assert len(step_proxies) == 1

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 2

    # TODO: For CWL - deactivate implicit scattering Galaxy does
    # and force annotation in the workflow of scattering? Maybe?
    wc_step = galaxy_workflow_dict["steps"][1]
    assert wc_step["input_connections"]

    assert "inputs" in wc_step
    wc_inputs = wc_step["inputs"]
    assert len(wc_inputs) == 1
    file_input = wc_inputs[0]
    assert file_input["scatter_type"] == "dotproduct", wc_step

    assert len(wc_step["workflow_outputs"]) == 1


def test_workflow_outputs_of_inputs():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/any-type-compat.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3
    input_step = galaxy_workflow_dict["steps"][0]

    assert len(input_step["workflow_outputs"]) == 1


def test_workflow_scatter_multiple_input():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines4-wf.cwl"))

    step_proxies = proxy.step_proxies()
    assert len(step_proxies) == 1

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3


def test_workflow_multiple_input_merge_flattened():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines7-wf.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3

    tool_step = galaxy_workflow_dict["steps"][2]
    assert "inputs" in tool_step
    inputs = tool_step["inputs"]
    assert len(inputs) == 1
    input = inputs[0]
    assert input["merge_type"] == "merge_flattened"


def test_workflow_step_value_from():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/step-valuefrom-wf.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3

    tool_step = [s for s in galaxy_workflow_dict["steps"].values() if s["label"] == "step1"][0]
    assert "inputs" in tool_step
    inputs = tool_step["inputs"]
    assert len(inputs) == 1
    assert "value_from" in inputs[0], inputs


def test_workflow_input_without_source():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/step-valuefrom3-wf.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3

    tool_step = galaxy_workflow_dict["steps"][2]
    assert "inputs" in tool_step
    inputs = tool_step["inputs"]
    assert len(inputs) == 3, inputs
    assert inputs[2].get("value_from")


def test_workflow_input_default():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/pass-unconnected.cwl"))
    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3

    tool_step = galaxy_workflow_dict["steps"][2]

    assert "inputs" in tool_step
    inputs = tool_step["inputs"]
    assert len(inputs) == 2, inputs
    assert inputs[1]


def test_search_workflow():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/search.cwl#main"))
    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 5


def test_workflow_simple_optional_input():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/io-int-optional-wf.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 2

    input_step = galaxy_workflow_dict["steps"][0]
    assert input_step["type"] == "parameter_input", input_step
    assert input_step["tool_state"]["parameter_type"] == "field", input_step


def test_boolean_defaults():
    proxy = workflow_proxy(_cwl_tool_path("v1.2/tests/conditionals/cond-wf-002_nojs.cwl"))
    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3
    bool_input = galaxy_workflow_dict["steps"][0]
    assert bool_input["label"] == "test", bool_input
    bool_tool_state = bool_input["tool_state"]
    assert bool_tool_state["optional"]
    assert bool_tool_state["default"]["value"] is False


def test_workflow_file_optional_input():
    proxy = workflow_proxy(_cwl_tool_path("v1.0/v1.0/count-lines11-wf.cwl"))

    galaxy_workflow_dict = proxy.to_dict()
    assert len(galaxy_workflow_dict["steps"]) == 3

    input_step = galaxy_workflow_dict["steps"][0]
    # TODO: make this File? - implemented in Galaxy now
    assert input_step["type"] == "parameter_input", input_step
    assert input_step["tool_state"]["optional"] is True, input_step


def test_load_proxy_simple():
    cat3 = _cwl_tool_path("v1.0/v1.0/cat3-tool.cwl")
    tool_source = get_tool_source(cat3)

    # Behavior was changed - too verbose?
    # description = tool_source.parse_description()
    # assert description == "Print the contents of a file to stdout using 'cat' running in a docker container.", description

    input_sources = _inputs(tool_source)
    assert len(input_sources) == 1

    input_source = input_sources[0]
    assert input_source.parse_help() == "The file that will be copied using 'cat'"
    assert input_source.parse_label() == "Input File"

    outputs, output_collections = tool_source.parse_outputs(None)
    assert len(outputs) == 1

    software_requirements, containers, resource_requirements, _, credentials = tool_source.parse_requirements()
    assert software_requirements.to_dict() == []
    assert len(containers) == 1
    assert containers[0].to_dict() == {
        "identifier": "debian:stretch-slim",
        "type": "docker",
        "resolve_dependencies": False,
        "shell": "/bin/sh",
    }
    assert len(resource_requirements) == 1
    assert resource_requirements[0].to_dict() == {"resource_type": "ram_min", "value_or_expression": 8}
    assert len(credentials) == 0


def test_representation_id():
    cat3 = _cwl_tool_path("v1.0/v1.0/cat3-tool.cwl")
    with open(cat3) as f:
        representation = yaml.safe_load(f)
        representation["id"] = "my-cool-id"

        uuid = str(uuid4())
        proxy = tool_proxy(tool_object=representation, uuid=uuid)
        tool_id = proxy.galaxy_id()
        assert tool_id == uuid, tool_id
        id_proxy = tool_proxy_from_persistent_representation(proxy.to_persistent_representation())
        tool_id = id_proxy.galaxy_id()
        assert tool_id == uuid, tool_id
        assert proxy.uuid == id_proxy.uuid


def test_env_tool1():
    env_tool1 = _cwl_tool_path("v1.0/v1.0/env-tool1.cwl")
    tool_source = get_tool_source(env_tool1)
    _inputs(tool_source)


def test_wc2_tool():
    env_tool1 = _cwl_tool_path("v1.0/v1.0/wc2-tool.cwl")
    tool_source = get_tool_source(env_tool1)
    _inputs(tool_source)
    datasets, collections = _outputs(tool_source)
    assert len(datasets) == 1, datasets
    output = datasets["output"]
    assert output.format == "expression.json", output.format


def test_optional_output():
    optional_output2_tool1 = _cwl_tool_path("v1.0/v1.0/optional-output.cwl")
    tool_source = get_tool_source(optional_output2_tool1)
    datasets, collections = _outputs(tool_source)
    assert len(datasets) == 2, datasets
    output = datasets["optional_file"]
    assert output.format == "expression.json", output.format


def test_schemadef_tool():
    tool_path = _cwl_tool_path("v1.0/v1.0/schemadef-tool.cwl")
    tool_source = get_tool_source(tool_path)
    _inputs(tool_source)


def test_params_tool():
    tool_path = _cwl_tool_path("v1.0/v1.0/params.cwl")
    tool_source = get_tool_source(tool_path)
    _inputs(tool_source)


def test_tool_reload():
    cat1_tool = _cwl_tool_path("v1.0/v1.0/cat1-testcli.cwl")
    tool_source = get_tool_source(cat1_tool)
    _inputs(tool_source)

    # Test reloading - had a regression where this broke down.
    cat1_tool_again = _cwl_tool_path("v1.0/v1.0/cat1-testcli.cwl")
    tool_source = get_tool_source(cat1_tool_again)
    _inputs(tool_source)


def _outputs(tool_source: CwlToolSource):
    return tool_source.parse_outputs(None)


def _inputs(tool_source=None, path=None):
    if tool_source is None:
        path = _cwl_tool_path(path)
        tool_source = get_tool_source(path)

    input_pages = tool_source.parse_input_pages()
    assert input_pages.inputs_defined
    page_sources = input_pages.page_sources
    assert len(page_sources) == 1
    page_source = page_sources[0]
    input_sources = page_source.parse_input_sources()
    return input_sources



import json
import os
import shutil
import tempfile

from galaxy.tool_util.toolbox.views.sources import StaticToolBoxViewSources
from galaxy.util import safe_makedirs
from .test_toolbox_view_parsing import EXAMPLE_1


def test_from_dicts():
    sources = StaticToolBoxViewSources(view_dicts=[EXAMPLE_1])
    defs = sources.get_definitions()
    assert len(defs) == 1
    assert defs[0].id == "rna"


def test_from_directory():
    target_dir = tempfile.mkdtemp()
    dir_1 = os.path.join(target_dir, "dir1")
    dir_2 = os.path.join(target_dir, "dir2")
    safe_makedirs(dir_1)
    safe_makedirs(dir_2)
    with open(os.path.join(dir_1, "foo.json"), "w") as f:
        json.dump(EXAMPLE_1, f)
    with open(os.path.join(dir_2, "seqanalysis.yml"), "w") as f:
        f.write("""
name: Sequence Analysis
type: generic
items:
- type: workflow
  id: 12345abcd
""")

    config_param = f"{os.path.abspath(dir_1)},{os.path.abspath(dir_2)}"
    try:
        sources = StaticToolBoxViewSources(view_directories=config_param)
        defs = sources.get_definitions()
        assert len(defs) == 2
        assert defs[0].id == "rna"
        assert defs[1].id == "seqanalysis"
    finally:
        shutil.rmtree(target_dir)



"""Test filters used by test_toolbox_filters.py."""


def filter_tool(context, tool):
    """Test Filter Tool"""
    return False


def filter_section(context, section):
    """Test Filter Section"""
    return False


def filter_label_1(context, label):
    """Test Filter Label 1"""
    return False


def filter_label_2(context, label):
    """Test Filter Label 2"""
    return False
